/////////////////////////////////////////////////////////////////////////////
// Name:        flipperb.h
// Purpose:     wxFlipperCtrlBase & twTabInfo class
// Author:      Hans Van Leemputten
// Modified by:
// Created:     01/May/2001
// RCS-ID:      $Id: flipperb.h,v 1.5 2001/07/18 01:16:15 georgetasker Exp $
// Copyright:   (c) Hans Van Leemputten & Aleksandras Gluchovas
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_FLIPPER_BASE_H_
#define _WX_FLIPPER_BASE_H_

#ifdef __GNUG__
    #pragma interface "flipperb.h"
#endif

// ----------------------------------------------------------------------------
// headers and constants
// ----------------------------------------------------------------------------
#include "wx/panel.h"

#define WXFlipperCTRLBASE_VERSION      1.0

// layout types for title bars of the tabs
// (are selected up by evaluating the available free space)

class twTabInfo; // forward decl.

#define wxTITLE_IMG_AND_TEXT 0
#define wxTITLE_IMG_ONLY     1
#define wxTITLE_BORDER_ONLY  2

// ----------------------------------------------------------------------------
// wxFlipperCtrlBase declaration
// ----------------------------------------------------------------------------

class wxFlipperCtrlBase : public wxPanel
{
    DECLARE_ABSTRACT_CLASS(wxFlipperCtrlBase)
    // DECLARE_EVENT_TABLE() => Declare these in the derived class!

public:
    // Constructor
    wxFlipperCtrlBase();
    wxFlipperCtrlBase( wxWindow *parent,
                       wxWindowID id = -1,
                       const wxPoint& pos = wxDefaultPosition,
                       const wxSize& size = wxDefaultSize,
                       long  style = 0,
                       const wxString& name = wxT("wxFlipperCtrlBase"));
    // Destructor
    virtual ~wxFlipperCtrlBase();

    // Create and init.
    bool Create( wxWindow *parent, wxWindowID id = -1,
                 const wxPoint& pos = wxDefaultPosition,
                 const wxSize& size = wxDefaultSize,
                 long  style = 0,
                 const wxString& name = wxT("wxFlipperCtrlBase"));

private:
    void DoInitFlipperCtrl(bool bCreated = FALSE);

public:
    // Get the number of Tabs controlled by this class.
    int GetTabCount();
    // Get the controlled object where tab eq to tabNo.
    wxWindow* GetTab(int tabNo);
    // Get the controlled object of the active tab
    wxWindow* GetActiveTab();
    // Is there a tab under control of this class?
    bool HasTab();
    // Set tabNo as active one.
    virtual void SetActiveTab(int tabNo);

    // Add a tab to the control.
    // tabs can be also added when the window is
    // already displayed - "on the fly"
    virtual void AddTab(wxWindow* pContent,             // contained window
                        wxString  tabText,              // tab label
                        int       tabID,                // tab ID Number  
                        wxString  imageFileName = "",   // if "", only text label is displayed
                        long      imageType     = wxBITMAP_TYPE_BMP);
    // NOTE:: if this AddTab(..) overload is called, the
    //        image bitmap will not be serialized (if performed),
    //        use the above method instead, so that images could
    //        be restored using the given file names
    virtual void AddTab(wxWindow* pContent,         // contained window
                        wxString  tabText,          // tab label
                        int       tabID,            // tab ID Number  
                        wxBitmap* pImage = NULL);
    // Remove a tab to the control.
    virtual void RemoveTab(int tabNo);
    // Is called when a Tab is added.
    virtual void OnTabAdded(twTabInfo* pInfo);

    // Get layout style.
    int GetLayoutStyle();
    // Set layout style.
    void SetLayoutStyle(int NewStyle = wxTITLE_IMG_AND_TEXT);

    // Refresh this object and the current active object...
    virtual void Refresh(bool eraseBackground = TRUE, const wxRect* rect = NULL);

    // Enable/Disable generation of a mask when adding a bitmap to a Tab
    void SetGenerateMasks(bool bOn);
    // Is mask generation Enable or Disable?
    bool GetGenerateMasks();
    // Set mask color, this color is used when adding a bitmap Tab.
    void SetMaskColor(const wxColour& color);
    // return the mask color in use.
    wxColour& GetMaskColor();
    
    /***************************/
    /* Functions to implement! */
    /***************************/

    // Should be invoked to restore the window without any changes
    // Like on paint...
    virtual void DrawDecorations(wxDC& dc) = 0;

    // Should be invoked to redisplay window with changed properties
    virtual void RecalcLayout(bool andRepaint = TRUE) = 0;

    /*******************************/
    /* End Functions to implement! */
    /*******************************/

    // overrride, to provide different font for tab-labels
    virtual wxFont GetLabelingFont();
    
protected:
    // Get tab info of the tab eq to tabNo 
    twTabInfo* GetTabInfo(int tabNo);
    // Get the tab info from the active tab.
    twTabInfo* GetActiveTabInfo();

    // Add a tab to the control.
    virtual void AddTab(twTabInfo *pTabInfo);

    // Get the coordinates where it is allowed to draw the title bar (Tab)
    virtual void GetTitleCoord(int *x1, int *y1, int *x2, int *y2);
    // Get the coordinates where it is allowed to draw the body in
    virtual void GetBodyCoord(int *x1, int *y1, int *x2, int *y2);

    // Re-Size and show or hide the body content.
    virtual void SizeBodyContent(int x,int y, int width, int height, bool repant);
    // Re-Size and show or hide the body content in a easy way.
    virtual void EasySizeBodyContent(bool repaint);

    // return -1, if non of the title bars was hitted,
    // otherwise the index of the hitted tab title bar
    virtual int HitTest(const wxPoint& pos);

    // Perpare the mask of a tab image.
    void PrepareBitmap(wxBitmap& bmp);

public:
    // event handlers
    void OnPaint(wxPaintEvent& event);
    void OnSize (wxSizeEvent& event);

protected:
    // Tab info list
    wxList m_Tabs;

    // State variables.
    int    m_ActiveTab;     // Activer tab number (Note: -1 if none)

    // State layout variables
    int    m_LayoutType;    // Specifies what to display as tab decoration.
                            // Possibilities: wxTITLE_IMG_AND_TEXT, wxTITLE_IMG_ONLY, wxTITLE_BORDER_ONLY
                            // Default: wxTITLE_IMG_AND_TEXT

    // Size/layout related variables (These don't have a Set/Get function).
    int    m_TitleHeight;   // This is the height reserved to draw the tab (or whatever)...
    int    m_VertGap;       // The Vertical gap between control border and new Ctrl,
                            // Default: 0
    int    m_HorizGap;      // The Horizontal gap between control border and new Ctrl,
                            // Default: 0
    
private:
    // Bitmap mask variables.
    bool m_GenerateMasks;
    wxColour m_MaskColor;
};

// ----------------------------------------------------------------------------
// twTabInfo declaration
// ----------------------------------------------------------------------------

// helper structure of wxFlipperCtrlBase
class twTabInfo : public wxObject
{
    DECLARE_DYNAMIC_CLASS(twTabInfo)

public:
    twTabInfo();
    ~twTabInfo();
    
    int ImgWidth();
    int ImgHeight();
    int ImageToTxtGap(int prefGap);
    
    bool HasImg();
    bool HasText();
    bool HandleHoriz();
    bool HandleVert();

    wxBitmap& GetImg();
    wxString& GetText();
    wxWindow& GetContent();
    
public: 
    int       m_TabID;      // tab ID Number
    wxWindow* m_pContent;   // tab content
    wxBitmap  m_BitMap;     // tab bitmap   
    wxString  m_Text;       // tab Text
    wxSize    m_Dims;       // tab size
    
    // used for serialization
    wxString  m_ImageFile;
    long      m_ImageType;

    // Used for handling scrollbars. (if any...)
    bool      m_HorizScrollBar;
    bool      m_VertScrollBar;
};

// ----------------------------------------------------------------------------
// Event declaration
// ----------------------------------------------------------------------------

enum PAGE_CONTROL_COMMAND_EVENTS
{
    wxEVT_TAB_RIGHT_CLICKED = 2000, //wxEVT_USER_FIRST,
    wxEVT_TAB_LEFT_CLICKED
};

#define EVT_TAB_RIGHT_CLICKED(id, fn) { wxEVT_TAB_RIGHT_CLICKED, id, -1, \
    (wxObjectEventFunction) (wxEventFunction) (wxCommandEventFunction) & fn, \
    (wxObject *) NULL },

#define EVT_TAB_LEFT_CLICKED(id, fn) { wxEVT_TAB_LEFT_CLICKED, id, -1, \
    (wxObjectEventFunction) (wxEventFunction) (wxCommandEventFunction) & fn, \
    (wxObject *) NULL },


#endif /* _WX_FLIPPER_BASE_H_ */

