///////////////////////////////////////////////////////////////////////////////
// Name:        espinbut.cpp
// Purpose:     Implementation of an enhanced version of wxSpinButton
// Author:      Hans Van Leemputten
// Modified by:
// Created:     21.01.2001
// RCS-ID:      $Id: espinbut.cpp,v 1.1 2001/09/07 15:07:39 hansvl Exp $
// Copyright:   Hans Van Leemputten & wxWorkshop Team (c) 2001
// Licence:     wxWindows licence
///////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
    #pragma implementation "espinbut.h"
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/bmpbuttn.h"
    #include "wx/timer.h"
    #include "wx/settings.h"
    #include "wx/dcmemory.h"
#endif

#include "wx/wspp/espinbut.h"

// ============================================================================
// declaration of wxBButtonHelper
// ============================================================================
class wxBButtonHelper : public wxBitmapButton
{
public:
    // Default constructor
    wxBButtonHelper()
        : wxBitmapButton()
    {
        m_pTimer     = NULL;
        m_bFirstTick = FALSE;
    }
    wxBButtonHelper(wxWindow *parent,
                   wxWindowID id,
                   const wxBitmap& bitmap,
                   const wxPoint& pos = wxDefaultPosition,
                   const wxSize& size = wxDefaultSize,
                   long style = wxBU_AUTODRAW,
                   const wxValidator& validator = wxDefaultValidator,
                   const wxString& name = wxT("wxBButtonHelper"))
        : wxBitmapButton(parent, id, bitmap, pos, size, style, validator, name)
    {
        m_bFirstTick = FALSE;
        m_pTimer     = new wxTimer( this );
        m_pTimer->Stop();        
    }

    // Destructor
    ~wxBButtonHelper()
    {
        wxDELETE(m_pTimer);
    }
    
    // Handle Events.
    void OnButtonUp(wxMouseEvent& event); 
    void OnButtonDown(wxMouseEvent& event);
    void OnProgressTimer(wxTimerEvent& event);

    // We need to overwrite this because when the button gets 
    // disabled we need to stop the timer!
    virtual bool Enable( bool enable = TRUE );

    // This function gives wxEnhSpinButton the option to stop the 
    // timer if it is  no longer needed to make time ticks. 
    // This function is introduced for wxGTK... because wxEVT_LEFT_UP is NOT always send!
    void StopTimer();
    
    // Send Event to base class...
    void ButtonMoreDown();

private:
    // Timer object
    wxTimer *m_pTimer;

    bool m_bFirstTick;

    DECLARE_DYNAMIC_CLASS(wxBButtonHelper)
    DECLARE_EVENT_TABLE()
};

// ============================================================================
// implementation of wxBButtonHelper
// ============================================================================

IMPLEMENT_DYNAMIC_CLASS( wxBButtonHelper, wxBitmapButton )

BEGIN_EVENT_TABLE( wxBButtonHelper, wxBitmapButton )
    EVT_LEFT_DOWN( wxBButtonHelper::OnButtonDown )
    EVT_LEFT_UP  ( wxBButtonHelper::OnButtonUp   )
    EVT_TIMER    ( -1, wxBButtonHelper::OnProgressTimer)
END_EVENT_TABLE()

void wxBButtonHelper::OnButtonUp(wxMouseEvent& event) 
{
    // Stop Timer
    StopTimer();
    
    // Change Focus back to the parrent class...
    if (GetParent())
        GetParent()->SetFocus();
    
    event.Skip();
}

void wxBButtonHelper::OnButtonDown(wxMouseEvent& event) 
{
    ButtonMoreDown();
    m_bFirstTick = TRUE;

    // Start the button timer.
    if (m_pTimer)
        m_pTimer->Start(150, FALSE);

    event.Skip();
}

void wxBButtonHelper::OnProgressTimer(wxTimerEvent& event)
{
    // We skip the first timer tick
    if (!m_bFirstTick)
        ButtonMoreDown();
    else
        m_bFirstTick = FALSE;
}

bool wxBButtonHelper::Enable(bool enable)
{
    // Stop Timer if the button is disabled
    if (!enable)
        StopTimer();

    // Call the base function
    return wxBitmapButton::Enable(enable);
}

void wxBButtonHelper::StopTimer()
{
    if (m_pTimer)
        m_pTimer->Stop();
}

void wxBButtonHelper::ButtonMoreDown()
{
    wxMouseEvent event(wxEVT_LEFT_DOWN);

    event.m_x = 0;  // For now it does not mather if it is 0
    event.m_y = 0;  // For now it does not mather if it is 0
    event.SetEventObject(this);

    if (GetParent())
        GetParent()->GetEventHandler()->ProcessEvent(event);
}

// ============================================================================
// implementation of wxEnhSpinButton
// ============================================================================

IMPLEMENT_DYNAMIC_CLASS(wxEnhSpinButton, wxControl)

BEGIN_EVENT_TABLE( wxEnhSpinButton, wxControl )
    EVT_LEFT_DOWN(wxEnhSpinButton::OnButton)
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// creation
// ----------------------------------------------------------------------------

wxEnhSpinButton::wxEnhSpinButton()
{
    Init();
}

wxEnhSpinButton::wxEnhSpinButton(wxWindow *parent,
                                 wxWindowID id,
                                 const wxPoint& pos,
                                 const wxSize& size,
                                 long style,
                                 const wxString& name)
{
    Init();

    (void)Create(parent, id, pos, size, style, name);
}

void wxEnhSpinButton::Init()
{
    m_value    = 0; // Current Value
    m_nButtons = 0; // Number of buttons created...

    // Init styles
    m_bMinMaxBnt = FALSE;
    m_bCenterBnt = FALSE;
    m_bArrowBnt  = FALSE;
    m_bDisableBnt= FALSE;

    // Init the button List.
    int i = 0;
    for (; i < 5; i++)
        m_pButton[i] = NULL;
}

wxEnhSpinButton::~wxEnhSpinButton()
{
    // TODO:: Add code here.
}

bool wxEnhSpinButton::Create(wxWindow *parent,
                             wxWindowID id,
                             const wxPoint& pos,
                             const wxSize& size,
                             long style,
                             const wxString& name)
{
    // the spin buttons never have the border
    style &= ~wxBORDER_MASK;

    // Create base class
    if ( !wxSpinButtonBase::Create(parent, id, pos, wxSize(50, 50), style | wxWANTS_CHARS  | wxNO_BORDER,
                                   wxDefaultValidator, name) )
        return FALSE;

    SetBackgroundColour( wxSystemSettings::GetSystemColour( wxSYS_COLOUR_BTNFACE ));

    // Check styles
    m_bMinMaxBnt = (wxSP_MINMAX_BUTTON & style) != 0;
    m_bCenterBnt = (wxSP_CENTER_BUTTON & style) != 0;
    m_bDisableBnt= (wxSP_DISABLE_BUTTON & style) != 0 && (wxSP_WRAP & style) == 0;
    m_bArrowBnt  = TRUE;

    /*
    m_bArrowBnt  = (wxSP_ARROW_KEYS & style) != 0;
    if (!m_bArrowBnt && (m_bMinMaxBnt || m_bCenterBnt))
        m_bArrowBnt = TRUE;
    */

    // 2 buttons is the minimum, Up and Down...
    m_nButtons = 2;

    // Add more buttons to the list if needed...
    if ( m_bMinMaxBnt )
        m_nButtons += 2;
    if ( m_bCenterBnt )
        m_nButtons += 1;
  
    // Construct all buttons...
    CreatButtons();

    SetBestSize(size);
    ChangeEnable();

    return TRUE;
}

void wxEnhSpinButton::CreatButtons()
{
    wxPoint arrow[3];           // List of Polygon points.
    bool    bDrawMinMaxButton;  // Is the current button a min/max button?
    wxCoord nExtra;             // 0 when bDrawMinMaxButton == FALSE, else wxWIDTH_MINMAX_LINE
    bool    bDrawCenterButton;  // Is the current button THE center button?

    int i;
    for(i = 0; i < m_nButtons ; i++)
    {
        // Size of the bitmap
        wxCoord  arrowWidth  = 0;
        wxCoord  arrowHeight = 0;

        // Set to default... We don't know yet.
        bDrawMinMaxButton = FALSE;
        nExtra            = 0;
        bDrawCenterButton = FALSE;
        
        if ( IsVertical() )
        {
            arrowWidth  = 7;
            arrowHeight = 4;

            // Check if it's a Min/Max Button?
            if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
            {
                bDrawMinMaxButton = TRUE;
                nExtra            = wxWIDTH_MINMAX_LINE;
                arrowHeight       += nExtra;
            }
            
            if (i < (m_nButtons/2))
            {
                /* Arrow Up         */
                // wxBitmap( 7, 4 );
                arrow[0] = wxPoint(0, 3 + nExtra);
                arrow[1] = wxPoint(3, 0 + nExtra);
                arrow[2] = wxPoint(6, 3 + nExtra);
            }
            else
            {
                /* Arrow Down       */
                // wxBitmap( 7, 4 );
                arrow[0] = wxPoint(0, 0);
                arrow[1] = wxPoint(3, 3);
                arrow[2] = wxPoint(6, 0);
            }

            // Check if it's the Center button?
            if ( m_bCenterBnt && (i == (m_nButtons/2)))
            {
                bDrawCenterButton = TRUE;

                arrowHeight = 2;
            }   
        }
        else
        {
            arrowWidth  = 4;
            arrowHeight = 7;

            // Check if it's a Min/Max Button?
            if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
            {
                bDrawMinMaxButton = TRUE;
                nExtra            = wxWIDTH_MINMAX_LINE
                arrowWidth       += nExtra;
            }

            if (i < (m_nButtons/2))
            {
                /* Arrow "<" Left   */
                // wxBitmap( 4, 7 );
                arrow[0] = wxPoint(3 + nExtra, 0);
                arrow[1] = wxPoint(0 + nExtra, 3);
                arrow[2] = wxPoint(3 + nExtra, 6);    
            }
            else
            {    
                /* Arrow ">" right  */
                // wxBitmap( 4, 7 );
                arrow[0] = wxPoint(0, 0);
                arrow[1] = wxPoint(3, 3);
                arrow[2] = wxPoint(0, 6);
            }

            // Check if it's the Center button?
            if ( m_bCenterBnt && (i == (m_nButtons/2)))
            {
                bDrawCenterButton = TRUE;

                arrowWidth = 2;
            }   
        }

        // Create arrow.
        wxBitmap bitmap(arrowWidth, arrowHeight);
        wxMemoryDC dc;
        dc.SelectObject( bitmap );
        if ( !bDrawCenterButton )
        {
            dc.SetPen(*wxBLACK_PEN);
            dc.SetBrush(*wxBLACK_BRUSH);
            dc.SetBackground(*wxWHITE_BRUSH);
            dc.Clear();
            dc.DrawPolygon(3, arrow);

            if ( bDrawMinMaxButton )
            {
                if ( IsVertical() )
                {
                    if (i < (m_nButtons/2))
                        dc.DrawLine( 0, 0, arrowWidth, 0);
                    else
                        dc.DrawLine( 0, arrowHeight-1, arrowWidth, arrowHeight-1);
                }
                else
                {
                    if (i < (m_nButtons/2))
                        dc.DrawLine( 0, 0, 0, arrowHeight);
                    else
                        dc.DrawLine( arrowWidth-1, 0, arrowWidth-1, arrowHeight);
                }
            }

            dc.SelectObject( wxNullBitmap );
    
            // Make the bitmap transparent.
            bitmap.SetMask(new wxMask(bitmap, *wxWHITE));
        }
        else
        {
            // This will draw a rectangle the size the bitmap... Like we want
            // Easy no?
            dc.SetBackground(*wxBLACK_BRUSH);
            dc.Clear();

            dc.SelectObject( wxNullBitmap );
        }
        

        // Create the button!
        m_pButton[i] = new wxBButtonHelper(this, -1, bitmap /*, wxDefaultPosition, wxSize(16, 16) */ );
    }    

}

void wxEnhSpinButton::DestroyButtons()
{
    int i;
    for (i = 0; i < m_nButtons; i++)
    {
        wxDELETE(m_pButton[i]);
    }
}

// ----------------------------------------------------------------------------
// size calculations
// ----------------------------------------------------------------------------

wxSize wxEnhSpinButton::DoGetBestSize() const
{
    wxSize bestSize(0, 0);

    // Hard code Size 16!!!
    if ( IsVertical() )
        bestSize.Set(16, m_nButtons * 16);
    else
        bestSize.Set(m_nButtons * 16, 16);
        
    return bestSize;
}

void wxEnhSpinButton::SetBestSize(const wxSize& size)
{
    // Set the best size
    wxSpinButtonBase::SetBestSize(size);

    // Fit and align all spin buttons.
    Fit();
}

void wxEnhSpinButton::Fit()
{
    // No need to waste cycles on something that does not exists.
    if (m_nButtons < 1) 
        return;

    // Get the size of the base widget...
    wxSize size = GetSize();

    // Now order the buttons...
    if ( IsVertical() )
    {
        wxCoord buttonSize = size.GetHeight() / m_nButtons;
        wxCoord x = 0;
        wxCoord y = 0;
        
        int i;
        for (i = 0; i < m_nButtons; i++)
        {
            m_pButton[i]->SetSize(x, y, size.GetWidth(), buttonSize);
            y += buttonSize;
        }
    }
    else
    {
        wxCoord buttonSize = size.GetWidth() / m_nButtons;
        wxCoord x = 0;
        wxCoord y = 0;
        
        int i;
        for (i = 0; i < m_nButtons; i++)
        {
            m_pButton[i]->SetSize(x, y+1, buttonSize, size.GetHeight()-1);
            x += buttonSize;
        }
    }
}

// ----------------------------------------------------------------------------
// value access
// ----------------------------------------------------------------------------

void wxEnhSpinButton::SetRange(int minVal, int maxVal)
{
    wxSpinButtonBase::SetRange(minVal, maxVal);

    // because the arrows disabled state might have changed - we don't check if
    // it really changed or not because SetRange() is called rarely enough and
    // son an extre refresh here doesn't really hurt
    Refresh();

    ChangeEnable();
}

int wxEnhSpinButton::GetValue() const
{
    return m_value;
}

void wxEnhSpinButton::SetValue(int val)
{
    if ( val != m_value )
    {
        m_value = val;

        Refresh();

        ChangeEnable();
    }
}

int wxEnhSpinButton::NormalizeValue(int value) const
{
    if ( value > m_max )
    {
        if ( GetWindowStyleFlag() & wxSP_WRAP )
            value = m_min + (value - m_max) % (m_max - m_min);
        else
            value = m_max;
    }
    else if ( value < m_min )
    {
        if ( GetWindowStyleFlag() & wxSP_WRAP )
            value = m_max - (m_min - value) % (m_max - m_min);
        else
            value = m_min;
    }

    return value;
}

bool wxEnhSpinButton::ChangeValue(int inc)
{
    int valueNew = NormalizeValue(m_value + inc);

    if ( valueNew == m_value )
    {
        // nothing changed - most likely because we are already at min/max
        // value
        return FALSE;
    }

    wxSpinEvent event(inc > 0 ? wxEVT_SCROLL_LINEUP : wxEVT_SCROLL_LINEDOWN,
                      GetId());
    event.SetPosition(valueNew);
    event.SetEventObject(this);

    if ( GetEventHandler()->ProcessEvent(event) && !event.IsAllowed() )
    {
        // programm has vetoed the event
        return FALSE;
    }

    m_value = valueNew;

    ChangeEnable();
    
    // send wxEVT_SCROLL_THUMBTRACK as well
    event.SetEventType(wxEVT_SCROLL_THUMBTRACK);
    (void)GetEventHandler()->ProcessEvent(event);

    return TRUE;
}

void wxEnhSpinButton::ChangeEnable()
{
    if (m_bDisableBnt)
    {
        // Enable all again...
        int i;
        for (i = 0; i < m_nButtons; i++)
        {
            m_pButton[i]->Enable(TRUE);
        }

        if ( IsVertical() )
        {   
            // Start disabling again...
            if (m_value == m_max)
            {
                m_pButton[0]->Enable(FALSE);
                
                if (m_bMinMaxBnt)
                    m_pButton[1]->Enable(FALSE);
            }
            
            if (m_value == m_min)
            {
                m_pButton[m_nButtons-1]->Enable(FALSE);
                
                if (m_bMinMaxBnt)
                    m_pButton[m_nButtons-2]->Enable(FALSE);
            }
        }
        else
        {
            // Start disabling again...
            if (m_value == m_max)
            {
                m_pButton[m_nButtons-1]->Enable(FALSE);
                
                if (m_bMinMaxBnt)
                    m_pButton[m_nButtons-2]->Enable(FALSE);
            }
            
            if (m_value == m_min)
            {
                m_pButton[0]->Enable(FALSE);
                
                if (m_bMinMaxBnt)
                    m_pButton[1]->Enable(FALSE);
            }
            
        }
    }
    else
    {
        if ((m_value == m_min || m_value == m_max) &&
            !(GetWindowStyleFlag() & wxSP_WRAP))
        {
            int i;
            for (i = 0; i < m_nButtons; i++)
            {
                m_pButton[i]->StopTimer();
            }
        }
    }
}

bool wxEnhSpinButton::Enable( bool enable )
{
    int i;
    for (i = 0; i < m_nButtons; i++)
    {
        m_pButton[i]->Enable(enable);
    }

    return TRUE;
}

// ----------------------------------------------------------------------------
// Button event handling.
// ----------------------------------------------------------------------------
void wxEnhSpinButton::OnButton( wxCommandEvent &event )
{
    int newValue = 0;
        
    int i;
    for (i = 0; i < m_nButtons; i++)
    {
        if (m_pButton[i] && m_pButton[i] == (wxBButtonHelper*)event.GetEventObject())
        {
            if ( IsVertical() )
            {
                if (i < (m_nButtons/2))
                {
                    newValue  = +1;

                    if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
                        newValue = m_max - m_value;
                }
                else
                {
                    newValue  = -1;

                    if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
                        newValue = m_min - m_value;
                }
            }
            else
            {
                if (i < (m_nButtons/2))
                {
                    newValue  = -1;

                    if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
                        newValue = m_min - m_value;
                }
                else
                {
                    newValue  = +1;

                    if ( m_bMinMaxBnt && ((i == 0) || (i == m_nButtons-1)) )
                        newValue = m_max - m_value;
                }
            }

            // Correct the value if it's the Center button.
            if ( m_bCenterBnt && (i == (m_nButtons/2)))
            {
                newValue = ((m_min + m_max) / 2) - m_value;
            }   

            break;
        }
    }

    // Determine inc value
    int inc = 0;
    if (newValue > 0)
        inc = +1;
    if (newValue < 0)
        inc = -1;

    // Change the spin value.
    for (i = abs(newValue); i > 0; i--)
        ChangeValue(inc);

    return;
}


